
#include "i2cdrv.h"
#include "FreeRTOS.h"
#include "queue.h"

#include "i2cdrv.h"
#include "efm32.h"
#include "dvk.h"
#include "semphr.h"

#include "efm32_gpio.h"

/*
 * Queue used to send item from interrupt routine to the I2CDRV_Transfer function
 *
 */
static xQueueHandle xRxedChars;

/*
 * Semaphore used to make sure only one task uses the I2C buss 
 * at once.
 */
static xSemaphoreHandle bsemI2C;

#include <stddef.h>
#include "dvk_boardcontrol.h"
#include "i2cdrv.h"
#include "efm32_cmu.h"
#include "efm32_gpio.h"

/***************************************************************************//**
 * @brief
 *   Initalize basic I2C master mode driver for use on the DVK.
 *
 * @details
 *   This driver only supports master mode, single bus-master. In addition
 *   to configuring the EFM32 I2C peripheral module, it also configures DVK
 *   specific setup in order to use the I2C bus.
 *
 * @param[in] init
 *   Pointer to I2C initialization structure.
 ******************************************************************************/
void I2CDRV_Init(const I2C_Init_TypeDef *init)
{
	int i;
		
	vSemaphoreCreateBinary(bsemI2C);
	xRxedChars = xQueueCreate( 1, ( unsigned portBASE_TYPE ) sizeof( signed portCHAR ) );
	
	DVK_enablePeripheral(DVK_I2C);
	
	CMU_ClockEnable(cmuClock_HFPER, true);
	CMU_ClockEnable(cmuClock_I2C0, true);
	
	/* Use location 3: SDA - Pin D14, SCL - Pin D15 */
	/* Output value must be set to 1 to not drive lines low... We set */
	/* SCL first, to ensure it is high before changing SDA. */
	GPIO_PinModeSet(gpioPortD, 15, gpioModeWiredAnd, 1);
	GPIO_PinModeSet(gpioPortD, 14, gpioModeWiredAnd, 1);
	
	/* In some situations (after a reset during an I2C transfer), the slave */
	/* device may be left in an unknown state. Send 9 clock pulses just in case. */
	for (i = 0; i < 9; i++)
	{
		/*
		 * TBD: Seems to be clocking at appr 80kHz-120kHz depending on compiler
		 * optimization when running at 14MHz. A bit high for standard mode devices,
		 * but DVK only has fast mode devices. Need however to add some time
		 * measurement in order to not be dependable on frequency and code executed.
		 */
		GPIO_PinModeSet(gpioPortD, 15, gpioModeWiredAnd, 0);
		GPIO_PinModeSet(gpioPortD, 15, gpioModeWiredAnd, 1);
	}
	
	/* Enable pins at location 3 (which is used on the DVK) */
	I2C0->ROUTE = I2C_ROUTE_SDAPEN |
				I2C_ROUTE_SCLPEN |
				(3 << _I2C_ROUTE_LOCATION_SHIFT);
	
	NVIC_SetPriority(I2C0_IRQn,7);
	/* Enable interrupt in Cortex Core */
	NVIC_ClearPendingIRQ(I2C0_IRQn);
	NVIC_EnableIRQ(I2C0_IRQn);
	
	I2C_Init(I2C0, init);
	CMU_ClockEnable(cmuClock_I2C0, false);	
}


/***************************************************************************//**
 * @brief
 *   Perform I2C transfer.
 *
 * @details
 *   This driver only supports master mode, single bus-master. It does not
 *   return until the transfer is complete. Uses interrupts to poll for completion
 *
 * @param[in] seq
 *   Pointer to sequence structure defining the I2C transfer to take place. The
 *   referenced structure must exist until the transfer has fully completed.
 ******************************************************************************/
I2C_TransferReturn_TypeDef I2CDRV_Transfer(I2C_TransferSeq_TypeDef *seq)
{
	I2C_TransferReturn_TypeDef ret;
	ret = I2C_TransferInit(I2C0, seq);
	if(ret == i2cTransferInProgress)
	{	
		xQueueReceive( xRxedChars, &ret, portMAX_DELAY );
	}
	return(ret);
}
/**************************************************************************//**
 * @brief I2C0 interrupt handler
 *****************************************************************************/
void I2C0_IRQHandler(void)
{
	I2C_TransferReturn_TypeDef ret;
	I2C_TypeDef *I2C = I2C0;
	portBASE_TYPE xHigherPriorityTaskWoken=pdFALSE;
	ret = I2C_Transfer(I2C0);
	if(ret != i2cTransferInProgress){
		xQueueSendFromISR(xRxedChars,&ret,&xHigherPriorityTaskWoken);
	}
	I2C->IFC = _I2C_IFC_MASK;
	portEND_SWITCHING_ISR(xHigherPriorityTaskWoken);
}

