/*Read this if LEUART and EM2 does not work!
Errata LEUART1:

LEUART + DMA
EM2 cannot be entered when transmit-
ting the last byte using LEUART and
DMA.

*/
#include "FreeRTOS.h"
#include "queue.h"
#include "task.h"
#include "semphr.h"

#include "efm32_cmu.h"
#include "efm32_gpio.h"
#include "efm32_leuart.h"

#include "serial.h"
#include "efm32.h"
#include "dvk.h"
#include "efm32_dma.h"

#define serialDMATXCHANEL 0
#define serialDMARXCHANEL 1
#define serialLEUART_SIGFRAME '\r'
/*
 * Semaphore used to Signal when the DMA cycle has completed.
 * Is released in the DMACallback() function.
 */
static xSemaphoreHandle bsemDmaTx;

/* 
 * Used to ensure that only one task tries to use the DMA for TX at a time.
 */
static xSemaphoreHandle bsemDmaActiveTx;

/*
 * Semaphore used to signal when the DMA receive cycle has completed.
 * Is released in the DMACallback() function 
 */
static xSemaphoreHandle bsemDmaRx;

/*
 * Called when a DMA cycle has completed.
 */
void DMACallback(unsigned int channel, bool primary, void *user);

/*
 * Structs used to init the DMA
 */
DMA_Init_TypeDef init;	
DMA_CB_TypeDef cbStruct;	
DMA_CfgChannel_TypeDef chCfgStuctTx;
DMA_CfgDescr_TypeDef cfgDescrTx;	
DMA_CfgChannel_TypeDef chCfgStuctRx;
DMA_CfgDescr_TypeDef cfgDescrRx;

/*
 * Two buffers are used when receiving from the LEUART using the DMA,
 * too ensure that the DMA does not write over something before we are
 * able to copy it.
 */
static char RxBuffer1[RXBUFFERSIZE];
static char RxBuffer2[RXBUFFERSIZE];
static char* RxBufferInUseByDMA = RxBuffer1;
static char* RxBufferIdle = RxBuffer2;

/**************************************************************************//**
 * @brief Intializes LEUART1 for use as an output interface, 9600-8-N-1
 *****************************************************************************/
static void prvLEUART1_Init(void)
{
  LEUART_TypeDef *leuart = LEUART1;
  LEUART_Init_TypeDef init;

  /* Enable CORE LE clock in order to access LE modules */
  CMU_ClockEnable(cmuClock_CORELE, true);

  /* Do not prescale clock */
  CMU_ClockDivSet(cmuClock_LEUART1, cmuClkDiv_1);

  /* Enable clock */
  CMU_ClockEnable(cmuClock_LEUART1, true);

  /* Use default location 0: TX - Pin C6, RX - Pin C7 */
  /* To avoid false start, configure output as high */
  GPIO_PinModeSet(gpioPortC, 6, gpioModePushPull, 1);
  /* Define input, no filtering */
  GPIO_PinModeSet(gpioPortC, 7, gpioModeInput, 0);

  /* Enable pins at default location */
  leuart->ROUTE = LEUART_ROUTE_RXPEN | LEUART_ROUTE_TXPEN;

  /* Configure LEUART1 */
  init.enable = leuartDisable;
  init.refFreq = 0;
  init.baudrate = 9600;
  init.databits = leuartDatabits8;
  init.parity = leuartNoParity;
  init.stopbits = leuartStopbits1;
  LEUART_Init(leuart, &init);

  /* Clear previous RX interrupts */
  LEUART_IntClear(LEUART1, LEUART_IF_SIGF);
  NVIC_ClearPendingIRQ(LEUART1_IRQn);
  LEUART1->CTRL |= LEUART_CTRL_TXDMAWU | LEUART_CTRL_RXDMAWU;  
  /* Enable RX interrupts */
  LEUART_IntEnable(LEUART1, LEUART_IEN_SIGF);
  NVIC_EnableIRQ(LEUART1_IRQn);
  while(LEUART1->SYNCBUSY);
  NVIC_SetPriority(LEUART1_IRQn,7);
  LEUART1->SIGFRAME = serialLEUART_SIGFRAME;
  
  /*If revision A of chip, TXC inerrupt and rxdatav must be active*/
  volatile uint32_t *reg = (volatile uint32_t *) 0xE00FFFE8;
  uint32_t minor = reg[0] & 0xF0;
  minor |= (reg[1] >> 4) & 0xF;
  if (minor == 0)
  {
        
	  LEUART1->IEN |=  LEUART_IEN_TXC |LEUART_IEN_RXDATAV;
  }
  
  /* Finally enable it */
  LEUART_Enable(leuart, leuartEnable);  
}


xComPortHandle xSerialPortInitMinimal( unsigned long ulWantedBaud, unsigned portBASE_TYPE uxQueueLength )
{ 
	vSemaphoreCreateBinary(bsemDmaTx);
	bsemDmaActiveTx = xSemaphoreCreateMutex();
	vSemaphoreCreateBinary(bsemDmaRx);
	xSemaphoreTake(bsemDmaTx,portMAX_DELAY);	
	xSemaphoreTake(bsemDmaRx,portMAX_DELAY);	
	
	CMU->HFPERCLKEN0 |= CMU_HFPERCLKEN0_USART2;
	DVK_enablePeripheral(DVK_RS232B);

	init.hprot = 0;
	char * tempPtr = pvPortMalloc((16 * DMA_CHAN_COUNT * 2 * 2));
	
	/*
	 *This if is needed to ensure that the pointer is alligned to 256 bits.
	 */
	if(((uint32_t)(tempPtr) & ((16 * DMA_CHAN_COUNT * 2) - 1))){
		tempPtr = (char *) ( ~((16 * DMA_CHAN_COUNT * 2) - 1)&(uint32_t)tempPtr);
		tempPtr += ((16 * DMA_CHAN_COUNT * 2));
	}
	init.controlBlock=(DMA_DESCRIPTOR_TypeDef *) tempPtr;
	DMA_Init(&init);

	chCfgStuctTx.highPri=1;
	chCfgStuctTx.enableInt = 1;
	chCfgStuctTx.select = DMAREQ_LEUART1_TXEMPTY;
	cbStruct = (DMA_CB_TypeDef){DMACallback,0,0};
	chCfgStuctTx.cb = &cbStruct; 
	DMA_CfgChannel(serialDMATXCHANEL,&chCfgStuctTx);

	cfgDescrTx.dstInc=dmaDataIncNone;
	cfgDescrTx.srcInc=dmaDataInc1;
	cfgDescrTx.size=dmaDataSize1;
	cfgDescrTx.arbRate=dmaArbitrate1;
	cfgDescrTx.hprot = 0;
  	DMA_CfgDescr(serialDMATXCHANEL,1,&cfgDescrTx);



	chCfgStuctRx.highPri=1;
	chCfgStuctRx.enableInt = 1;
	chCfgStuctRx.select = DMAREQ_LEUART1_RXDATAV;
	chCfgStuctRx.cb = &cbStruct; 
	DMA_CfgChannel(serialDMARXCHANEL,&chCfgStuctRx);

	cfgDescrRx.dstInc=dmaDataInc1;
	cfgDescrRx.srcInc=dmaDataIncNone;
	cfgDescrRx.size=dmaDataSize1;
	cfgDescrRx.arbRate=dmaArbitrate1;
	cfgDescrRx.hprot = 0;
  	DMA_CfgDescr(serialDMARXCHANEL,1,&cfgDescrRx);

	/*Important to set the DMA interrupt to a level below configMAX_SYSCALL_INTERRUPT_PRIORITY*/
	NVIC_SetPriority(DMA_IRQn,7);
	prvLEUART1_Init();

	DMA_ActivateBasic(serialDMARXCHANEL,1,0,RxBufferInUseByDMA,(void *)&LEUART1->RXDATA,RXBUFFERSIZE-1);
	
	return NULL;
};

/*
 * pcString is the string to send. 
 * The function will not return untill the string has been transfered by DMA,
 * hence it is safe to reuse the pointer.
 */
void vSerialPutString(char * pcString, unsigned short usStringLength ){
        xSemaphoreTake(bsemDmaActiveTx,portMAX_DELAY); //ensures that only task is allowed to start the DMA at a time.
        DMA_ActivateBasic(serialDMATXCHANEL,1,0,(void *)&LEUART1->TXDATA,pcString,usStringLength-1);
	xSemaphoreTake(bsemDmaTx,portMAX_DELAY); //wait untill the DMA cycle is complete.
	xSemaphoreGive(bsemDmaActiveTx);
}

/*
 * Returns, in pcRxString the string read. The pointer needs
 * to have room for RXBUFFERSIZE bytes.
 */
void xSerialGetString( char * pcRxedString )
{
	xSemaphoreTake(bsemDmaRx,portMAX_DELAY);
	int i = 0;
	while(RxBufferIdle[i] != serialLEUART_SIGFRAME & i < RXBUFFERSIZE){
		pcRxedString[i] = RxBufferIdle[i];
		i++;
	}
	pcRxedString[i] = '\0';//insert a \0 at the end to make it a propper string.
}

void DMACallback(unsigned int channel, bool primary, void *user){
    static signed portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;  
	if(channel==serialDMATXCHANEL){
          xSemaphoreGiveFromISR(bsemDmaTx, &xHigherPriorityTaskWoken );
	}
	if(channel==serialDMARXCHANEL){
		static char * tmp;  
		tmp = RxBufferInUseByDMA;
		RxBufferInUseByDMA= RxBufferIdle;
		RxBufferIdle = tmp;
		DMA_ActivateBasic(serialDMARXCHANEL,1,0,RxBufferInUseByDMA,(void *)&LEUART1->RXDATA,RXBUFFERSIZE-1);		
		xSemaphoreGiveFromISR(bsemDmaRx, &xHigherPriorityTaskWoken );
	}
    portEND_SWITCHING_ISR( xHigherPriorityTaskWoken );	
}

/*
 * Only the SIGFRAME interrupt is enabled.
 * Meaning that this interrut occurs when the LEUART recives a character
 * defined in the serialLEUART_SIGFRAME macro. 
 *
 * When a sigframe arrives, stop the DMA channel. and run the DMACallback function.
 */
void LEUART1_IRQHandler(){
        int tmp;
	if(LEUART1->IF & LEUART_IF_SIGF){
		DMA->CHENC = 0x1UL << serialDMARXCHANEL;  
		LEUART1->IFC=LEUART_IF_SIGF;	
		DMACallback(serialDMARXCHANEL,1,NULL);
	}
	if(LEUART1->IF & LEUART_IF_TXC ){
		LEUART1->IFC = LEUART_IF_TXC;		
	}else if(LEUART1->IF &LEUART_IF_RXDATAV){
		tmp = LEUART1->RXDATA;
	}
}
